"""
Wisdom is the user interaction with Whatify wisdom, it contains

"""
# Prediction is the way to productize the Ensemble created in the previous steps. Once an Ensemble is created,
# users can upload additional Datasources that may be used for predictions.
#
# ‘Prediction’ API includes querying of predictions (Get, List and Delete) and creating a Prediction to get predictions
# on existing Ensembles and uploaded Datasources.
import os
from io import StringIO
import boto3
import pandas as pd
import s3fs
import toolkit_w
from toolkit_w.internal.api_requestor import APIRequestor
from toolkit_w.internal.whatify_response import WhatifyResponse
from toolkit_w.resources.api_resource import APIResource


class Connector(APIResource):
    _CLASS_PREFIX = 'connectors'
    credentials = None
    s3_path = None
    s3_client = None
    s3_resource = None

    @classmethod
    def get_credentials(cls, user_token: str = None, renew: bool = False):
        if (cls.credentials is None) or renew:
            url = '/'.join([cls._CLASS_PREFIX, 'whatify_connect_permissions'])
            requester = APIRequestor()
            if user_token is None:
                user_token = toolkit_w.token
            cls.credentials = requester.get(url, params={'jwt': user_token})
        return cls.credentials

    @classmethod
    def get_s3_path(cls, user_token: str = None):
        cls.get_credentials(user_token)
        if cls.s3_path is None:
            bucket = cls.credentials['bucket']
            PATH = cls.credentials['path']
            cls.s3_path = f's3://{bucket}/{PATH}/'
        return cls.s3_path

    @classmethod
    def get_client_path(cls, path: bool = False, user_token: str = None):
        file_list = []
        cls.get_credentials(user_token)
        os.environ['AWS_ACCESS_KEY_ID'] = cls.credentials['access_key']
        os.environ['AWS_SECRET_ACCESS_KEY'] = cls.credentials['secret_key']
        os.environ['AWS_SESSION_TOKEN'] = cls.credentials['session_token']
        os.environ['AWS_DEFAULT_REGION'] = cls.credentials['region']
        fs = s3fs.S3FileSystem()
        try:
            file_list = fs.ls(cls.get_s3_path(user_token=user_token), refresh=True)
            if not path:
                if len(file_list) > 0:
                    # remove first item - path location
                    pre_path = file_list.pop(0)
                    # remove path from list
                    file_list = {x.replace(str(pre_path), '') for x in file_list}

        except PermissionError as ex:
            print(' '.join(['folder is empty:', str(ex)]))
        return file_list

    @classmethod
    def get_s3_boto_client(cls, user_token: str = None):
        cls.get_credentials(user_token)
        cls.s3_client = boto3.client('s3',
                                     aws_access_key_id=cls.credentials['access_key'],
                                     aws_secret_access_key=cls.credentials['secret_key'],
                                     aws_session_token=cls.credentials['session_token'])
        return cls.s3_client

    @classmethod
    def get_s3_boto_resource(cls, user_token: str = None):
        cls.get_credentials(user_token)
        cls.s3_resource = boto3.resource('s3',
                                         aws_access_key_id=cls.credentials['access_key'],
                                         aws_secret_access_key=cls.credentials['secret_key'],
                                         aws_session_token=cls.credentials['session_token'])
        return cls.s3_resource

    @classmethod
    def read_to_pandas_df(cls, file_name: str, user_token: str = None, **kwargs):
        cls.get_s3_boto_client(user_token=user_token)
        obj = cls.s3_client.get_object(Bucket=cls.credentials['bucket'], Key='/'.join([cls.credentials['path'], file_name]))
        df = pd.read_csv(obj['Body'], **kwargs)
        return df

    @classmethod
    def write_pandas_df_to_s3(cls, df, file_name: str, user_token: str = None, **kwargs):
        cls.get_s3_boto_resource(user_token=user_token)
        csv_buffer = StringIO()
        df.to_csv(csv_buffer, **kwargs)
        cls.s3_resource.Object(cls.credentials['bucket'], '/'.join([cls.credentials['path'], file_name])).put(Body=csv_buffer.getvalue())



import os
from collections import OrderedDict

import requests
import uuid

import toolkit_w
from toolkit_w.internal.errors import AuthenticationError, APIError, InvalidRequestError, APIConnectionError, PermissionError
from toolkit_w.internal.whatify_response import WhatifyResponse


class APIRequestor(object):
    def __init__(self, http_client=None):
        if http_client is None:
            self._http_client = requests
        else:
            self._http_client = http_client

    def parse_filter_parameters(self, filter):
        if filter:
            filters = []
            for field, values in filter.items():
                for value in values:
                    filters.append("{}:{}".format(field, value))
            return filters

    def parse_sort_parameters(self, sort):
        assert sort is None or isinstance(sort, OrderedDict)
        if sort:
            sorts = []
            for field, value in sort.items():
                sorts.append('{}:{}'.format(field, value))
            return sorts

    def request(self, method, url, headers=None, body=None, params=None, api_key=None):
        if method not in ['GET', 'POST', 'PUT', 'DELETE', 'PATCH']:
            raise APIConnectionError(
                "Unrecognized HTTP method {method}. This may indicate a bug in the Whatify "
                "internal bindings. Please contact support for assistance.".format(method=method)
            )

        if api_key is None:
            token = self._get_token()
        else:
            token = api_key

        rheaders = self._build_headers()
        rheaders.update(**(headers or {}))

        params = {'jwt': token, **(params or {})}
        abs_url = "{base_url}/{url}".format(base_url=toolkit_w.api_base, url=url)
        response = self._http_client.request(method=method, url=abs_url, headers=rheaders, json=body, params=params)
        return self._handle_response(response)

    def post(self, url, headers=None, body=None, params=None, api_key=None):
        return self.request("POST", url, headers, body, params, api_key)

    def patch(self, url, headers=None, body=None, params=None, api_key=None):
        return self.request("PATCH", url, headers, body, params, api_key)

    def get(self, url, headers=None, body=None, params=None, api_key=None):
        return self.request("GET", url, headers, body, params, api_key)

    def delete(self, url, headers=None, body=None, params=None, api_key=None):
        return self.request("DELETE", url, headers, body, params, api_key)

    def put(self, url, headers=None, body=None, params=None, api_key=None):
        return self.request("PUT", url, headers, body, params, api_key)

    def _build_headers(self):
        return {'X-Request-ID': str(uuid.uuid4())}

    def _handle_response(self, response):
        response_json = {}
        try:
            response_json = response.json()
        except ValueError:
            pass
        if 200 <= response.status_code < 300:
            return self._handle_ok(response, response_json)
        else:
            if 400 <= response.status_code < 500 and response_json:
                raise self._handled(response)
            else:
                raise self._unhandled(response)

    def _handle_ok(self, response, response_json):
        if not response_json:
            return WhatifyResponse(headers=response.headers, status_code=response.status_code)

        if 'result' not in response_json:
            response_json = {'result': response_json}

        response_type = type(response_json['result'])
        if response_type == dict:
            result = WhatifyResponse(data=response_json.get('result', response_json), headers=response.headers,
                                     status_code=response.status_code)
        elif response_type == bool:
            result = WhatifyResponse(data=response_json, headers=response.headers,
                                     status_code=response.status_code)
        elif response_type == int:
            result = WhatifyResponse(data={'id': response_json['result']}, headers=response.headers,
                                     status_code=response.status_code)
        else:
            result = WhatifyResponse(data=response_json, headers=response.headers,
                                     status_code=response.status_code)
        return result

    def _handled(self, response):
        response_json = response.json()
        if response.status_code in [400, 404]:
            err = InvalidRequestError(message=response_json['error'], headers=response.headers,
                                      code=response.status_code, json_body=response_json)
        elif response.status_code == 401:
            err = AuthenticationError(message="Invalid token.", headers=response.headers,
                                      code=response.status_code, json_body=response_json)
        elif response.status_code == 403:
            err = PermissionError(message=response_json['message'], headers=response.headers, code=response.status_code,
                                  json_body=response_json)
        else:
            err = APIError(message=response_json['message'], headers=response.headers, code=response.status_code,
                           json_body=response_json)
        return err

    def _unhandled(self, response):
        try:
            raise APIError(response.json().get('error') or response.json().get('message'))
        except ValueError:
            raise APIError('API problem exception during request.')

    def _get_token(self):
        if toolkit_w.token is None:
            toolkit_w.token = os.getenv("FIREFLY_TOKEN", None)
            if toolkit_w.token is None:
                raise AuthenticationError("No token found. Please use `Whatify.login()` to create a token,"
                                          "or use `TOKEN` environment variable to manually use one."
                                          "If problem persists, please contact support.")
        return toolkit_w.token
